//
//  VirtualKBLayer.swift
//
//  Copyright (C) 2016 Hiroshi Kanemoto.
//	Released under the MIT license
//	http://opensource.org/licenses/mit-license.php
//
//	 Created:	2016/04/09	H.Kanemoto
//	Modified:
//
import Foundation
import UIKit

open class VirtualKBLayer : CALayer {
	fileprivate var parent_bound:CGRect!
	fileprivate var initial_rect:CGRect!
	fileprivate var cell_width:CGFloat!
	fileprivate var axis_x_even:[CGFloat] = [CGFloat]()
	fileprivate var axis_x_odd:[CGFloat] = [CGFloat]()
	fileprivate var axis_y:[CGFloat] = [CGFloat]()
	fileprivate var fontpt:CGFloat = 12
	//-------------------------------------------------------------------------
	//	コンストラクタ
	//-------------------------------------------------------------------------
	public init(rct:CGRect) {
		super.init()
		
		self.parent_bound = rct
		self.backgroundColor = UIColor.black.cgColor
		self.opacity = 0.5
		self.setNeedsDisplay()
		self.initial_rect = self.calcFrame(rct)
		self.setFrameSilent(self.initial_rect)
	}

	public override init(layer: Any) {
		super.init(layer: layer)
	}
	
	required public init?(coder aDecoder: NSCoder) {
		fatalError("init(coder:) has not been implemented")
	}
	//-------------------------------------------------------------------------
	//	デストラクタ
	//-------------------------------------------------------------------------
	deinit {
		self.axis_x_even.removeAll()
		self.axis_x_odd.removeAll()
		self.axis_y.removeAll()
	}
	//-------------------------------------------------------------------------
	//	座標処理
	//-------------------------------------------------------------------------
	func calcFrame(_ parent:CGRect) -> CGRect {
		self.axis_x_even.removeAll()
		self.axis_x_odd.removeAll()
		self.axis_y.removeAll()
		
		let w = parent.width * 0.9
		let ax = w / 32
		let cell = w / 16
		self.cell_width = cell
		self.fontpt = cell / 3
		let h = cell * 6
		let x = (parent.width - w) / 2
		let y = parent.height - h
		let rect = CGRect(x: x, y: y, width: w, height: h)
		for x:Int in 0 ..< 16 {
			axis_x_even.append(CGFloat(x) * cell)
			axis_x_odd.append(CGFloat(x) * cell + ax)
		}
		for y:Int in 0 ..< 6 {
			axis_y.append(CGFloat(y) * cell)
		}
		self.prepareKeys()
		return rect
	}
	func calcCellFrame(_ cell:Int, row:Int, block:Int) -> CGRect {
		let odd:Bool = row == 2 || row == 4 || row == 5
		let x = odd ? self.axis_x_odd[cell] : self.axis_x_even[cell]
		let y = self.axis_y[row]
		var rct = CGRect(x: x, y: y, width: self.cell_width * CGFloat(block), height: self.cell_width)
		rct = rct.insetBy(dx: self.cell_width * 0.07, dy: self.cell_width * 0.07)
		return rct
	}
	//-------------------------------------------------------------------------
	//	描画処理
	//-------------------------------------------------------------------------
	open override func draw(in ctx: CGContext) {
		ctx.setFillColor(UIColor.white.cgColor)
		ctx.setStrokeColor(UIColor.white.cgColor)
		ctx.setTextDrawingMode(CGTextDrawingMode.fillStroke)
		ctx.setLineWidth(1)
		let appDelegate:AppDelegate = UIApplication.shared.delegate as! AppDelegate
		for k in appDelegate.keys {
			if k.Rect.width > 0 && k.Rect.height > 0 {
				self.drawKey(ctx, key: k)
			}
		}
		
		ctx.translateBy(x: 0, y: self.bounds.height)
		ctx.scaleBy(x: 1, y: -1)
		for k in appDelegate.keys {
			if k.Rect.width > 0 && k.Rect.height > 0 {
				self.drawKeyLabel(ctx, key: k)
			}
		}
	}
	
	func drawKey(_ ctx: CGContext, key:N80Key) {
		let b = UIBezierPath(roundedRect: key.Rect, byRoundingCorners: UIRectCorner.allCorners, cornerRadii: CGSize(width: 5.0, height: 5.0))
		ctx.addPath(b.cgPath)
		if key.Pushed {
			ctx.fillPath()
		} else {
			ctx.strokePath()
		}
	}
	
	func drawKeyLabel(_ ctx: CGContext, key:N80Key) {
		let attr = NSAttributedString(string: key.Lable, attributes: [
			NSAttributedString.Key.font: UIFont(name:"Arial", size:self.fontpt)!,
			NSAttributedString.Key.foregroundColor: key.Pushed ? UIColor.black : UIColor.white])
		/*let attr = NSAttributedString(string: key.Lable, attributes: [
			NSFontAttributeName: UIFont(name:"HiraMinProN-W6", size:18)!,
			NSForegroundColorAttributeName: UIColor.whiteColor()])*/
		let line = CTLineCreateWithAttributedString(attr)
		let x = key.Rect.origin.x + key.Rect.width / 2 - (CGFloat(key.Lable.lengthOfBytes(using: String.Encoding.shiftJIS)) * self.fontpt / 2)
		//CGContextSetTextPosition(ctx, x, self.frame.height - key.Rect.origin.y - key.Rect.height / 2)
		ctx.textPosition = CGPoint(x: x, y: self.frame.height - key.Rect.origin.y - key.Rect.height / 2)
		CTLineDraw(line, ctx)
	}
	//-------------------------------------------------------------------------
	//	Key当たり処理
	//-------------------------------------------------------------------------
	open func hitKeyTest(_ pt:CGPoint) -> N80Key? {
		let key = self.hitKeyTextCore(pt)
		if key != nil && !key!.Pushed {
			self.setNeedsDisplay(self.bounds)
			return key
		}
		return nil
	}
	
	func hitKeyTextCore(_ pt:CGPoint) -> N80Key? {
		let appDelegate:AppDelegate = UIApplication.shared.delegate as! AppDelegate
		if pt.y < self.axis_y[1] {
			return self.hitTestCol(pt, arr: appDelegate.keys0)
		}
		if pt.y < self.axis_y[2] {
			return self.hitTestCol(pt, arr: appDelegate.keys1)
		}
		if pt.y < self.axis_y[3] {
			return self.hitTestCol(pt, arr: appDelegate.keys2)
		}
		if pt.y < self.axis_y[4] {
			return self.hitTestCol(pt, arr: appDelegate.keys3)
		}
		if pt.y < self.axis_y[5] {
			return self.hitTestCol(pt, arr: appDelegate.keys4)
		}
		return self.hitTestCol(pt, arr: appDelegate.keys5)
	}
	
	func hitTestCol(_ pt:CGPoint, arr:[N80Key]) -> N80Key? {
		for k in arr {
			if k.Rect.contains(pt) {
				return k
			}
		}
		return nil
	}
	//-------------------------------------------------------------------------
	//	Layer移動処理
	//-------------------------------------------------------------------------
	open func toggleShow() -> Bool {
		if self.frame.width < 1 {
			self.frame = self.initial_rect
			self.setNeedsDisplay(self.bounds)
			return true		//show
		} else {
			let rct = CGRect(x: 0, y: 0, width: 0, height: 0)
			self.frame = rct
			return false	//hide
		}
	}
	//-------------------------------------------------------------------------
	//	Layer移動処理
	//-------------------------------------------------------------------------
	open func setFrameSilent(_ rct:CGRect) {
		CATransaction.begin()
		CATransaction.setValue(kCFBooleanTrue, forKey: kCATransactionDisableActions)
		self.frame = rct
		CATransaction.commit()
	}
	func prepareKeys() {
		let appDelegate:AppDelegate = UIApplication.shared.delegate as! AppDelegate
		
		appDelegate.keys0[0].Rect = self.calcCellFrame(0, row: 0, block: 1)
        appDelegate.keys0[1].Rect = self.calcCellFrame(1, row: 0, block: 1)
		appDelegate.keys0[2].Rect = self.calcCellFrame(2, row: 0, block: 1)
		appDelegate.keys0[3].Rect = self.calcCellFrame(3, row: 0, block: 2)
		appDelegate.keys0[4].Rect = self.calcCellFrame(5, row: 0, block: 2)
		appDelegate.keys0[5].Rect = self.calcCellFrame(7, row: 0, block: 2)
		appDelegate.keys0[6].Rect = self.calcCellFrame(9, row: 0, block: 2)
		appDelegate.keys0[7].Rect = self.calcCellFrame(11, row: 0, block: 2)
		appDelegate.keys0[8].Rect = self.calcCellFrame(13, row: 0, block: 1)
		appDelegate.keys0[9].Rect = self.calcCellFrame(15, row: 0, block: 1)
		
		appDelegate.keys1[0].Rect = self.calcCellFrame(1, row: 1, block: 1)
		appDelegate.keys1[1].Rect = self.calcCellFrame(2, row: 1, block: 1)
		appDelegate.keys1[2].Rect = self.calcCellFrame(3, row: 1, block: 1)
		appDelegate.keys1[3].Rect = self.calcCellFrame(4, row: 1, block: 1)
		appDelegate.keys1[4].Rect = self.calcCellFrame(5, row: 1, block: 1)
		appDelegate.keys1[5].Rect = self.calcCellFrame(6, row: 1, block: 1)
		appDelegate.keys1[6].Rect = self.calcCellFrame(7, row: 1, block: 1)
		appDelegate.keys1[7].Rect = self.calcCellFrame(8, row: 1, block: 1)
		appDelegate.keys1[8].Rect = self.calcCellFrame(9, row: 1, block: 1)
		appDelegate.keys1[9].Rect = self.calcCellFrame(10, row: 1, block: 1)
		appDelegate.keys1[10].Rect = self.calcCellFrame(11, row: 1, block: 1)
		appDelegate.keys1[11].Rect = self.calcCellFrame(12, row: 1, block: 1)
		appDelegate.keys1[12].Rect = self.calcCellFrame(13, row: 1, block: 1)
		appDelegate.keys1[13].Rect = self.calcCellFrame(14, row: 1, block: 1)
		appDelegate.keys1[14].Rect = self.calcCellFrame(15, row: 1, block: 1)
		
		appDelegate.keys2[0].Rect = self.calcCellFrame(0, row: 2, block: 1)
		appDelegate.keys2[1].Rect = self.calcCellFrame(1, row: 2, block: 1)
		appDelegate.keys2[2].Rect = self.calcCellFrame(2, row: 2, block: 1)
		appDelegate.keys2[3].Rect = self.calcCellFrame(3, row: 2, block: 1)
		appDelegate.keys2[4].Rect = self.calcCellFrame(4, row: 2, block: 1)
		appDelegate.keys2[5].Rect = self.calcCellFrame(5, row: 2, block: 1)
		appDelegate.keys2[6].Rect = self.calcCellFrame(6, row: 2, block: 1)
		appDelegate.keys2[7].Rect = self.calcCellFrame(7, row: 2, block: 1)
		appDelegate.keys2[8].Rect = self.calcCellFrame(8, row: 2, block: 1)
		appDelegate.keys2[9].Rect = self.calcCellFrame(9, row: 2, block: 1)
		appDelegate.keys2[10].Rect = self.calcCellFrame(10, row: 2, block: 1)
		appDelegate.keys2[11].Rect = self.calcCellFrame(11, row: 2, block: 1)
		appDelegate.keys2[12].Rect = self.calcCellFrame(12, row: 2, block: 1)
		appDelegate.keys2[13].Rect = self.calcCellFrame(13, row: 2, block: 2)
		
		appDelegate.keys3[0].Rect = self.calcCellFrame(1, row: 3, block: 1)
		appDelegate.keys3[1].Rect = self.calcCellFrame(2, row: 3, block: 1)
		appDelegate.keys3[2].Rect = self.calcCellFrame(3, row: 3, block: 1)
		appDelegate.keys3[3].Rect = self.calcCellFrame(4, row: 3, block: 1)
		appDelegate.keys3[4].Rect = self.calcCellFrame(5, row: 3, block: 1)
		appDelegate.keys3[5].Rect = self.calcCellFrame(6, row: 3, block: 1)
		appDelegate.keys3[6].Rect = self.calcCellFrame(7, row: 3, block: 1)
		appDelegate.keys3[7].Rect = self.calcCellFrame(8, row: 3, block: 1)
		appDelegate.keys3[8].Rect = self.calcCellFrame(9, row: 3, block: 1)
		appDelegate.keys3[9].Rect = self.calcCellFrame(10, row: 3, block: 1)
		appDelegate.keys3[10].Rect = self.calcCellFrame(11, row: 3, block: 1)
		appDelegate.keys3[11].Rect = self.calcCellFrame(12, row: 3, block: 1)
		appDelegate.keys3[12].Rect = self.calcCellFrame(13, row: 3, block: 1)
		appDelegate.keys3[13].Rect = self.calcCellFrame(14, row: 3, block: 1)
		
		appDelegate.keys4[0].Rect = self.calcCellFrame(0, row: 4, block: 2)
		appDelegate.keys4[1].Rect = self.calcCellFrame(2, row: 4, block: 1)
		appDelegate.keys4[2].Rect = self.calcCellFrame(3, row: 4, block: 1)
		appDelegate.keys4[3].Rect = self.calcCellFrame(4, row: 4, block: 1)
		appDelegate.keys4[4].Rect = self.calcCellFrame(5, row: 4, block: 1)
		appDelegate.keys4[5].Rect = self.calcCellFrame(6, row: 4, block: 1)
		appDelegate.keys4[6].Rect = self.calcCellFrame(7, row: 4, block: 1)
		appDelegate.keys4[7].Rect = self.calcCellFrame(8, row: 4, block: 1)
		appDelegate.keys4[8].Rect = self.calcCellFrame(9, row: 4, block: 1)
		appDelegate.keys4[9].Rect = self.calcCellFrame(10, row: 4, block: 1)
		appDelegate.keys4[10].Rect = self.calcCellFrame(11, row: 4, block: 1)
		appDelegate.keys4[11].Rect = self.calcCellFrame(12, row: 4, block: 1)
		appDelegate.keys4[12].Rect = self.calcCellFrame(13, row: 4, block: 2)
		
		appDelegate.keys5[0].Rect = self.calcCellFrame(2, row: 5, block: 1)
		appDelegate.keys5[1].Rect = self.calcCellFrame(3, row: 5, block: 7)
		appDelegate.keys5[2].Rect = self.calcCellFrame(11, row: 5, block: 1)
		appDelegate.keys5[3].Rect = self.calcCellFrame(12, row: 5, block: 1)
		appDelegate.keys5[4].Rect = self.calcCellFrame(14, row: 5, block: 1)
	}
}
