//
//  geometry.swift
//
//  Copyright (C) 2015-2016 Hiroshi Kanemoto.
//	Released under the MIT license
//	http://opensource.org/licenses/mit-license.php
//
//	 Created:	2015/06/30	H.Kanemoto
//	Modified:
//
import Foundation

open class geometry: NSObject {
    class func toDegree(_ radian:Double) -> Double {
        return radian * 180.0 / Double.pi
    }
    class func toRadian(_ degree:Double) -> Double {
        return degree * Double.pi / 180.0
    }
    // 座標pの原点からの距離を返す
    class func distance(_ p:Position) -> Double {
        return self.distance(p.X, y: p.Y)
    }
    // 座標pの原点からの距離を返す
    class func distance(_ x:Double, y:Double) -> Double {
        return sqrt(pow(x, 2) + pow(y, 2))
    }
    // 線分abとcdが交差するかをチェックする
    class func intersection(_ ab:Line, cd:Line) -> Bool {
        return intersection(ab.Pos1, p2: ab.Pos2, p3: cd.Pos1, p4: cd.Pos2)
    }
    // 線分p1-p2とp3-p4が交差するかをチェックする
    class func intersection(_ p1:Position, p2:Position, p3:Position, p4:Position) -> Bool {
        // X座標によるチェック
        if p1.X >= p2.X {
            if (p1.X < p3.X && p1.X < p4.X) || (p2.X > p3.X && p2.X > p4.X) {
                return false
            }
        } else {
            if (p2.X < p3.X && p2.X < p4.X) || (p1.X > p3.X && p1.X > p4.X) {
                return false
            }
        }
        // Y座標によるチェック
        if p1.Y >= p2.Y {
            if (p1.Y < p3.Y && p1.Y < p4.Y) || (p2.Y > p3.Y && p2.Y > p4.Y) {
                return false
            }
        } else {
            if (p2.Y < p3.Y && p2.Y < p4.Y) || (p1.Y > p3.Y && p1.Y > p4.Y) {
                return false
            }
        }
        if ((p1.X - p2.X) * (p3.Y - p1.Y) + (p1.Y - p2.Y) * (p1.X - p3.X)) *
           ((p1.X - p2.X) * (p4.Y - p1.Y) + (p1.Y - p2.Y) * (p1.X - p4.X)) > 0.0 {
            return false
        }
        if ((p3.X - p4.X) * (p1.Y - p3.Y) + (p3.Y - p4.Y) * (p3.X - p1.X)) *
           ((p3.X - p4.X) * (p2.Y - p3.Y) + (p3.Y - p4.Y) * (p3.X - p2.X)) > 0.0 {
            return false
        }
        return true
    }
    //回転
    class func rotateLine(_ line:Line, degree:Double) -> Line {
        let a = rotate(line.Pos1, degree: degree)
        let b = rotate(line.Pos2, degree: degree)
        return Line(a: a, b: b)
    }
    class func rotate(_ p:Position, degree:Double) -> Position {
        return geometry.rotate(p.X, y:p.Y, degree:degree)
    }
    // |x'|   |cos -sin||x|
    // |  | = |        || |
    // |y'|   |sin  cos||y|
    //
    // x' = x*cos - y*sin
    // y' = x*sin + y*cos
    //
    // 回転は反時計周りなので、時計周りの角度を反転させる
    class func rotate(_ x:Double, y:Double, degree:Double) -> Position {
        let rad = toRadian(360 - degree)
        let sinx = sin(rad)
        let cosx = cos(rad)
        return Position(x:(x*cosx - y*sinx), y:(x*sinx + y*cosx))
    }
    class func scale(_ target:Position, factor:Double) -> Position {
        return Position(x: target.X * factor, y: target.Y * factor)
    }
    class func circumference(_ r:Double, degree:Double) -> Position {
        let rad = toRadian(degree)
        return Position(x: cos(rad)*r, y: sin(rad)*r)
    }
    class func circumcircleL(_ r:Double, n:Int) -> Double {
        // 正n角形の外接円半径より一辺の長さ(a)を返す
        //            a
        //  r = --------------
        //       2sin(pi/n)
        return sin(Double.pi / Double(n)) * 2 * r
    }
    class func incircleR(_ a:Double, n:Int) -> Double {
        // 一辺の長さ(a)の正n角形の内接円の半径を返す
        //            a
        //  r = --------------
        //       2tan(pi/n)
        return a / (tan(Double.pi / Double(n)) * 2)
    }
    class func diagonal(_ rect:CGRect) -> Double {
        // 対角線の交差点の座標がpの矩形を作成する
       return distance(Double(rect.width), y: Double(rect.height))
    }
    class func rectAtPosition(_ p:Position, width:Double, height:Double) -> CGRect {
        //対角線の交差点の座標がpの矩形を作成する
        return CGRect(x: p.X - (width / 2), y: p.Y + (height / 2), width: width, height: height)
    }
    class func rotativeDirection(_ a:Vector, b:Vector) -> Int {
        //		y
        //		|
        //		|      →  P'
        //		|      b ／
        //		|      ／φ
        //		|     O--------P
        //		|         →
        //		|         a
        //		+------------------→x
        //		S ≡ (Px - Ox) * (P'y - Oy) - (Py - Oy) * (P'x - Ox)
        //		S > 0 --- 左周り
        //		S < 0 --- 右周り
        //		S = 0 --- 一直線上
        let s = a.X * b.Y - a.Y * b.X
        if s == 0 {
            return 0
        } else if s > 0 {
            return -1
        } else {
            return 1
        }
    }
    class func innerProduct(_ a:Vector, b:Vector) -> Double {
        // A ・ B ≡ |A| * |B| * cosθ
        // A × B ≡ |A| * |B| * sinθ
        // A ・ B ≡ Ax * Bx + Ay * By
        // A × B ≡ Ax * By - Ay * Bx
        // θ ＝ atan2(A×B，A・B)
        let ab = a.X * b.X + a.Y * b.Y
        let axb = a.X * b.Y - a.Y * b.X
        let degree = atan2(axb, ab)
        return toDegree(degree)
    }
}
