//
//  SMBOpen.m
//
//  Copyright (C) 2016 Hiroshi Kanemoto.
//	Released under the MIT license
//	http://opensource.org/licenses/mit-license.php
//
//	 Created:	2016/03/23	H.Kanemoto
//	Modified:	2017/05/24	H.Kanemoto  SMB Signature
//
#import <Foundation/Foundation.h>
#import "SMBOpen.h"
#import "SMB2QueryInfo.h"
#import "SMB2Read.h"
#import "SMB2Close.h"
/*----------------------------------------------------------------------*
 *	SMB OPEN
 *----------------------------------------------------------------------*/
@implementation SMBOpen
/*----------------------------------------*
 *	Init/Dealloc
 *----------------------------------------*/
-(id)initWith:(SMBTreeConnect *)_treeconn FileId:(UInt8[16])_fileid AutoDisconnect:(BOOL)autodisconnect {
	if (self = [super init]) {
		self->treeconn = _treeconn;
		self->treeid = _treeconn.TreeId;
		self.Session = _treeconn.Session;
		memcpy(self->fileid, _fileid, 16);
		self->autotreedisconnect = autodisconnect;
		self->opened = false;
	}
	return self;
}
-(void)dealloc {
	DLog(@"");
	[self close];
	self->treeconn = nil;
}
/*----------------------------------------*
 *	Properties
 *----------------------------------------*/
-(SMBTreeConnect*)getTreeConnect {
	return self->treeconn;
}
-(TcpClient*)getTcp {
	return self.TreeConnect.Session.Connection.Tcp;
}
/*----------------------------------------*
 *	CLOSE
 *----------------------------------------*/
-(void)close {
	SMB2ReqClose *req = [[SMB2ReqClose alloc] initWith:self->treeid FileId:self->fileid];
	[self.TreeConnect.Session MarkRequest:req];
	if (req == nil || ![self.Session sendRequest:req])
		return;
	SMB2ResClose *res = [[SMB2ResClose alloc] init];
	if (res == nil || ![res recvResponse:self.getTcp] || res.Header->Status != 0)
		return;
	if (self->autotreedisconnect)
		[self->treeconn Disconnect];
}
/*----------------------------------------*
 *	READ
 *----------------------------------------*/
-(void)read:(SMBReadProgressCallback)callback {
	TBErrors code = TB_NO_ERR;
	code = [self read_core:callback];
	if (code != TB_NO_ERR) {
		if (callback != nil) {
			NSError *_err = [SMBUtils buildErr:code];
			callback(nil, 0, 0, true, _err);
		}
	}
}
-(TBErrors)read_core:(SMBReadProgressCallback)callback {
	int64_t endoffile = 0;
	for (;;) {
		//Query Info request
		SMB2ReqQueryInfo *req = [SMB2ReqQueryInfo create:self->treeid FileId:self->fileid];
		[self.TreeConnect.Session MarkRequest:req];
		if (req == nil || ![self.Session sendRequest:req])
			break;

		//Query Info response
		SMB2ResQueryInfo *res = [[SMB2ResQueryInfo alloc] init];
		if (res == nil || ![res recvResponse:self.getTcp])
			break;
		if (res.Header->Status == 0x103) {
			if (![res recvResponse:self.getTcp])
				break;
		}
		if (res.Header->Status != 0)
			break;
		
		//get information
		[res.Coder moveptrto:PACKETHEADERSIZE+QUERYINFORESSIZE-1];
		FileStandardInformation *info = [[FileStandardInformation alloc] init];
		if (info == nil)
			break;
		if (![info deserialize:res.Coder])
			break;
		endoffile = info.Info->EndOfFile;
		
		//do read file
		return [self readloop:callback EndOfFile:endoffile];
	}
	return SMB2_QUERY_INFO_REQ_ERR;
}
-(TBErrors)readloop:(SMBReadProgressCallback)callback EndOfFile:(int64_t)endoffile {
	TBErrors result = SMB2_READ_ERR;
	int64_t offset = 0, readlen = 0;
	
	SMB2ReqRead *req = [[SMB2ReqRead alloc] init];
	SMB2ResRead *res = [[SMB2ResRead alloc] initWith:(int32_t)self.TreeConnect.Session.ReadUnitSize];
	if (req == nil || res == nil)
		return result;
	[req.Coder consume:1];	//emputy payload then consume Buffer[1]
	req.Header->s.Sync.TreeId = self->treeid;
	req.Header->CreditRequest = 256;
	memcpy(req.READ->FileId, self->fileid, 16);
	req.READ->MinimumCount = 0;
	req.READ->RemainingBytes = 0;
	
	for (;;) {
		int64_t remain = endoffile - offset;
		int64_t len = self.TreeConnect.Session.ReadUnitSize;
		if (remain < len)
			len = remain;
		if (remain > 0x7fffffff)
			remain = 0x7fffffff;	//Max 32bit integer value
		
		//READ request
		req.READ->Length = (int)len;
		req.READ->Offset = offset;
		req.Header->CreditCharge = (int16_t)(1 + (((int)len - 1) / 65536));
		[self.TreeConnect.Session MarkRequest:req];
		if (![self.Session sendRequest:req])
			break;
		
		//READ response
		if (![res recvResponse:self.getTcp])
			break;
		if (res.Header->Status != 0)
			break;
		
		offset += res.READ->DataLength;
		readlen += res.READ->DataLength;
		BOOL finished = readlen >= endoffile;
		//callback to delive read data to caller
		if (callback != nil) {
			NSData *data = [[NSData alloc] initWithBytes:&res.Coder.Data[res.READ->DataOffset] length:res.READ->DataLength];
			callback(data, readlen, endoffile, finished, nil);
		}
		//return if finish to read
		if (finished) {
			result = TB_NO_ERR;
			break;
		}
	}
	return result;
}
/*----------------------------------------*
 *	READ ONCE
 *----------------------------------------*/
-(bytecoder*)read_once {
	int64_t endoffile = 0;
	for (;;) {
		//Query Info request
		SMB2ReqQueryInfo *req = [SMB2ReqQueryInfo create:self->treeid FileId:self->fileid];
		[self.TreeConnect.Session MarkRequest:req];
		if (req == nil || ![self.Session sendRequest:req])
			break;
		
		//Query Info response
		SMB2ResQueryInfo *res = [[SMB2ResQueryInfo alloc] init];
		if (res == nil || ![res recvResponse:self.getTcp])
			break;
		if (res.Header->Status == 0x103) {
			if (![res recvResponse:self.getTcp])
				break;
		}
		if (res.Header->Status != 0)
			break;
		
		//get information
		[res.Coder moveptrto:PACKETHEADERSIZE+QUERYINFORESSIZE-1];
		FileStandardInformation *info = [[FileStandardInformation alloc] init];
		if (info == nil)
			break;
		if (![info deserialize:res.Coder])
			break;
		endoffile = info.Info->EndOfFile;
		bytecoder *data = [[bytecoder alloc] init];
		if (data == nil)
			break;
		if (![data alloc_buffer:(int)endoffile])
			break;
		
		//do read file
		if ([self readloop_once_core:data EndOfFile:endoffile] == TB_NO_ERR)
			return data;
		[data freeptr];
		break;
	}
	return nil;
}
-(TBErrors)readloop_once_core:(bytecoder*)data EndOfFile:(int64_t)endoffile {
	TBErrors result = SMB2_READ_ERR;
	int64_t offset = 0, readlen = 0;
	
	SMB2ReqRead *req = [[SMB2ReqRead alloc] init];
	SMB2ResRead *res = [[SMB2ResRead alloc] initWith:(int32_t)self.TreeConnect.Session.ReadUnitSize];
	if (req == nil || res == nil)
		return result;
	[req.Coder consume:1];	//emputy payload then consume Buffer[1]
	req.Header->s.Sync.TreeId = self->treeid;
	req.Header->CreditRequest = 256;
	memcpy(req.READ->FileId, self->fileid, 16);
	req.READ->MinimumCount = 0;
	req.READ->RemainingBytes = 0;
	
	for (;;) {
		int64_t remain = endoffile - offset;
		int64_t len = self.TreeConnect.Session.ReadUnitSize;
		if (remain < len)
			len = remain;
		if (remain > 0x7fffffff)
			remain = 0x7fffffff;	//Max 32bit integer value
		
		//READ request
		req.READ->Length = (int)len;
		req.READ->Offset = offset;
		req.Header->CreditCharge = (int16_t)(1 + (((int)len - 1) / 65536));
		[self.TreeConnect.Session MarkRequest:req];
		if (![self.Session sendRequest:req])
			break;
		
		//READ response
		if (![res recvResponse:self.getTcp])
			break;
		if (res.Header->Status != 0)
			break;
		
		offset += res.READ->DataLength;
		readlen += res.READ->DataLength;
		BOOL finished = readlen >= endoffile;
		//callback to delive read data to caller
		if (![data push:&res.Coder.Data[res.READ->DataOffset] Length:res.READ->DataLength])
			break;
		//return if finish to read
		if (finished) {
			result = TB_NO_ERR;
			break;
		}
	}
	return result;
}

@end
