//
//  bytecoder.m
//
//  Copyright (C) 2016 Hiroshi Kanemoto.
//	Released under the MIT license
//	http://opensource.org/licenses/mit-license.php
//
//	 Created:	2016/03/13	H.Kanemoto
//	Modified:
//
#import <Foundation/Foundation.h>
#import "bytecoder.h"
#import "SMBUtils.h"
#include <malloc/malloc.h>
/*----------------------------------------------------------------------*
 *	Data buffer decoder/encoder
 *----------------------------------------------------------------------*/
@implementation bytecoder
/*----------------------------------------*
 *	Init/Dealloc
 *----------------------------------------*/
-(id)init {
	if (self = [super init]) {
		self->ptr = nil;
		self->siz_alloc = 0;
		self->siz_used = 0;
		self->siz_remain = 0;
		self->local_offset = 0;
		self->top_offset = 0;
	}
	return self;
}

-(void)dealloc {
	DLog(@"");
	[self freeptr];
}
/*----------------------------------------*
 *	Properties
 *----------------------------------------*/
-(UInt8 *)getData {
	return self->ptr;
}
-(void *)getPointer {
	return &self.Data[self->local_offset];
}
-(int)getLenght {
	return self->siz_used;
}
-(int)getTopOffset {
	return self->top_offset;
}
-(void)setTopOffset:(int)TopOffset {
	self->top_offset = TopOffset;
}
-(int)getGlobalOffset {
	return self->top_offset + self->local_offset;
}
-(int)getLocalOffset {
	return self->local_offset;
}
-(int)getOffset {
	return self->local_offset;
}
-(void)setOffset:(int)Offset {
	self->local_offset = Offset;
}
-(int)getUnreadCnt {
	int unread = self->siz_used - self->local_offset;
	return unread > 0 ? unread : 0;
}
-(int)getRemain {
	return self->siz_remain;
}
/*----------------------------------------*
 *	Data Buffer Operation
 *----------------------------------------*/
-(BOOL)alloc_buffer:(int)initial_size {
	void *p = calloc(initial_size, 1);
	if (p != nil) {
		self->siz_alloc	= initial_size;
		self->siz_used = 0;
		self->siz_remain = initial_size;
		self->ptr = p;
		return true;
	}
	return false;
}

-(BOOL)expand_buffer:(int)expand_size {
	if (expand_size > 0) {
		if (self->ptr == nil)
			return [self alloc_buffer:expand_size];
		void *p = realloc(self->ptr, self->siz_alloc + expand_size);
		if (p != nil) {
			self->siz_alloc += expand_size;
			self->siz_remain += expand_size;
			self->ptr = p;
			return true;
		}
	}
	return false;
}

-(BOOL)recycle_buffer:(int)recycle_size {
	if (recycle_size > 0) {
		if (self->ptr == nil){
			return [self alloc_buffer:recycle_size];
		} else if (self->siz_alloc >= recycle_size) {
			[self rewind];
			return true;
		} else {
			[self freeptr];
			if ([self alloc_buffer:recycle_size])
				return true;
		}
	}
	return false;
}

-(void)freeptr {
	if (self->ptr != nil) {
		free(self->ptr);
		assert(malloc_zone_check(NULL));
		self->ptr = nil;
	}
}

-(bytecoder *)cutout:(uint32_t)offset Length:(uint32_t)len; {
	bytecoder *p = [[bytecoder alloc] init];
	if (p != nil) {
		if ([p alloc_buffer:len]) {
			[p push:&self.Data[offset] Length:len];
			[p moveptrto:0];
			p.TopOffset = self.TopOffset + offset;
		}
	}
	return p;
}
/*----------------------------------------*
 *	Write Pointer Operation
 *----------------------------------------*/
-(void)consume:(int)siz {
	if (siz > 0) {
		self->siz_remain -= siz;
		self->siz_used += siz;
		self->local_offset += siz;
	}
}

-(void)rewind {
	self->siz_remain = self->siz_alloc;
	self->siz_used = 0;
	self->local_offset = 0;
}
/*----------------------------------------*
 *	Read Pointer Operation
 *----------------------------------------*/
-(BOOL)moveptr:(int)siz {
	int abs = self->local_offset + siz;
	return [self moveptrto:abs];
}

-(BOOL)moveptrto:(int)abs {
	if (abs < 0 || abs > self->siz_used)
		return false;
	self->local_offset = abs;
	return true;
}
-(BOOL)moveglobalptrto:(int)globalabs {
	int offset = globalabs - self->top_offset;
	return [self moveptrto:offset];
}
/*----------------------------------------*
 *	Write Access
 *----------------------------------------*/
-(BOOL)push:(const UInt8 *)buffer Length:(int)len {
	if (len >= 0) {
		if (self->siz_remain < len) {
			uint32_t lack = len - self->siz_remain;
			if (![self expand_buffer:lack])
				return false;
		}
		memcpy(&self.Data[self->siz_used], buffer, len);
		[self consume:len];
		return true;
	}
	return false;
}

-(BOOL)push_byte:(UInt8)data {
	UInt8 buffer[1];
	buffer[0] = data;
	return [self push:buffer Length:1];
}

-(BOOL)push_le_short:(UInt16)value {
	UInt8 buffer[2];
	[SMBUtils Short2ByteLE:value :buffer];
	return [self push:buffer Length:2];
}

-(BOOL)push_le_int:(UInt32)value {
	UInt8 buffer[4];
	[SMBUtils Int2ByteLE:value :buffer];
	return [self push:buffer Length:4];
}

-(BOOL)push_le_long:(UInt64)value {
	UInt8 buffer[8];
	[SMBUtils Long2ByteLE:value :buffer];
	return [self push:buffer Length:8];
}

-(BOOL)pad4bytes_align:(BOOL)global {
	int diff = [self diff4bytes_align:global];
	if (diff > 0) {
		UInt8 padding[diff];
		memset(padding, 0, diff);
		return [self push:padding Length:diff];
	}
	return true;
}

-(BOOL)pad8bytes_align:(BOOL)global {
	int diff = [self diff8bytes_align:global];
	if (diff > 0) {
		UInt8 padding[diff];
		memset(padding, 0, diff);
		return [self push:padding Length:diff];
	}
	return true;
}
/*----------------------------------------*
 *	Read Access
 *----------------------------------------*/
-(BOOL)pull:(UInt8 *)buffer Length:(int)len {
	if (self.UnreadCnt >= len) {
		memcpy(buffer, self.Pointer, len);
		[self moveptr:len];
		return true;
	}
	return false;
}

-(BOOL)pull_byte:(UInt8*)buffer {
	return [self pull:buffer Length:1];
}

-(BOOL)pull_le_short:(UInt16*)value {
	UInt8 buffer[2];
	if ([self pull:buffer Length:2]) {
		*value = [SMBUtils Byte2ShortLE:buffer];
		return true;
	}
	return false;
}

-(BOOL)pull_le_int:(UInt32*)value {
	UInt8 buffer[4];
	if ([self pull:buffer Length:4]) {
		*value = [SMBUtils Byte2IntLE:buffer];
		return true;
	}
	return false;
}

-(BOOL)pull_le_long:(UInt64*)value {
	UInt8 buffer[8];
	if ([self pull:buffer Length:8]) {
		*value = [SMBUtils Byte2LongLE:buffer];
		return true;
	}
	return false;
}

-(int)skip4bytes_align:(BOOL)global {
	int diff = [self diff4bytes_align:global];
	[self moveptr:diff];
	return diff;
}

-(int)skip8bytes_align:(BOOL)global {
	int diff = [self diff8bytes_align:global];
	[self moveptr:diff];
	return diff;
}
/*----------------------------------------*
 *	Misc
 *----------------------------------------*/
-(int)diff4bytes_align:(BOOL)global {
	uint32_t offset = global ? self.GlobalOffset : self.LocalOffset;
	uint32_t align = [SMBUtils Align4bytes:offset];
	int diff = align - offset;
	return diff;
}

-(int)diff8bytes_align:(BOOL)global {
	uint32_t offset = global ? self.GlobalOffset : self.LocalOffset;
	uint32_t align = [SMBUtils Align8bytes:offset];
	int diff = align - offset;
	return diff;
}

@end
/*----------------------------------------------------------------------*
 *	Packet base
 *----------------------------------------------------------------------*/
@implementation packetbase

-(id)init {
	if (self = [super init]) {
		self->coder = [[bytecoder alloc] init];
	}
	return self;
}
-(id)initWith:(bytecoder *)_coder {
	if (self = [super init]) {
		self->coder = _coder;
	}
	return self;
}

-(bytecoder *)getCoder {
	return self->coder;
}

-(bytecoder*)detach {
	bytecoder *b = self.Coder;
	self->coder = nil;
	return b;
}

@end
